/*
    execute.cpp: contains fucntions for accessing input data from xml config file and passing them to the vision functions. 
    Copyright (C) <2015>  <Safoura Rezapour Lakani>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
		
		In the case of questions, please contact "safoura.rezapour-lakani@uibk.ac.at"
*/
#include "execute.h"

void execute::read_input()
{
  TiXmlDocument doc("../config-data.xml");
  bool loadOkay = doc.LoadFile();
  TiXmlHandle hdoc(&doc);
  if( loadOkay )
  { 
    cout << "succesfully loaded" << endl;
    TiXmlElement* elem = hdoc.FirstChild("config").FirstChild("dims").ToElement();
    if( elem ){
      cout << "elem dims found" << endl;
      string val_s = elem->GetText();
      stringstream ss(val_s);
      int val;
      ss >> val;
      cout << val << endl;
      exec_mode_.dims = val;
    }
    else
      cout << "elem dims not found" << endl;
    TiXmlElement* elem1 = hdoc.FirstChild("config").FirstChild("patch-cluster").ToElement();
    if( elem1 )
    {
      string val = elem1->GetText();
      cout << "patch-cluster: " << val << endl;
      exec_mode_.patch_cls_path = val;
    }
   
    TiXmlElement* elem3 = hdoc.FirstChild("config").FirstChild("mode").ToElement();
    if( elem3 )
    {
      string val = elem3->GetText();
      cout << "mode:" << val << endl;
      if( mode_map_.find(val) != mode_map_.end() ){
        exec_mode_.mode = mode_map_[val];
        cout << "mode map is: " << mode_map_[val] << endl;
      }
    }
 
    TiXmlElement* elem4 = hdoc.FirstChild("config").FirstChild("prototypes").ToElement();
    if( elem4 )
    {
      string val = elem4->GetText();
      cout << "prototypes:" << val << "." << endl;
      exec_mode_.pro_path = val;
    }

    TiXmlElement* elem8 = hdoc.FirstChild("config").FirstChild("patch-cluster-mat").ToElement();
    if( elem8 )
    {
      string val = elem8->GetText();
      cout << "patch-cluster-mat: " << val << endl;
      exec_mode_.patch_cls_mat_path = val;
    }

    TiXmlElement* elem9 = hdoc.FirstChild("config").FirstChild("eval-path").ToElement();
    if( elem9 )
    {
      string val = elem9->GetText();
      cout << "eval-path: " << val << endl;
      exec_mode_.eval_path = val;
    }

    TiXmlElement* elem11 = hdoc.FirstChild("config").FirstChild("input-pcd").ToElement();
    if( elem11 )
    {
      string val = elem11->GetText();
      cout << "input-pcd: " << val << endl;
      exec_mode_.input_pcd = val;
    }

    TiXmlElement* elem14 = hdoc.FirstChild("config").FirstChild("input-dir").ToElement();
    if( elem14 )
    {
      string val = elem14->GetText();
      cout << "input-dir: " << val << endl;
      exec_mode_.input_dir = val;
    }

    TiXmlElement* elem16 = hdoc.FirstChild("config").FirstChild("default-path").ToElement();
    if( elem16 )
    {
      string val = elem16->GetText();
      cout << "default-path: " << val << endl;
      default_path_ = val;
		}

    TiXmlElement* elem17 = hdoc.FirstChild("config").FirstChild("voxel-resolution").ToElement();
    if( elem17 ){
      string val_s = elem17->GetText();
      stringstream ss(val_s);
      double val;
      ss >> val;
      voxel_resolution = val;
    }

    TiXmlElement* elem18 = hdoc.FirstChild("config").FirstChild("seed-resolution").ToElement();
    if( elem18 ){
      string val_s = elem18->GetText();
      stringstream ss(val_s);
      double val;
      ss >> val;
      seed_resolution = val;
    }

    TiXmlElement* elem19 = hdoc.FirstChild("config").FirstChild("patch-n-mat").ToElement();
    if( elem19 ){
      string val_s = elem19->GetText();
      exec_mode_.patch_n_cls_mat = val_s;
    }
    TiXmlElement* elem20 = hdoc.FirstChild("config").FirstChild("debug-mode").ToElement();
    if( elem20 ){
      string val_s = elem20->GetText();
      stringstream ss(val_s);
      bool val;
      ss >> val;
      debug_ = val;
    }
    TiXmlElement* elem21 = hdoc.FirstChild("config").FirstChild("sample-perc").ToElement();
    if( elem21 ){
      string val_s = elem21->GetText();
      stringstream ss(val_s);
      double val;
      ss >> val;
      sample_perc_ = val;
    }
  }
  else
  {
    cout << "load not succesful" << endl;
  }
}

void execute::execution()
{
  mode_map_["sample"] = SAMPLE;
  mode_map_["load"] = LOAD;
  mode_map_["parts"] = PARTS;  
  mode_map_["parts_all"] = PARTS_ALL;
  mode_map_["cluster"] = CLUSTER;
  mode_map_["evaluate"] = EVALUATE;
	mode_map_["part_seg"] = PART_SEG;
  mode_map_["estimate_region_models"] = ESTIMATE_REGION_MODELS;
	mode_map_["estimate_patch_threshold"] = ESTIMATE_PATCH_THRESHOLD;
	mode_map_["collect_statistics"] = COLLECT_STATISTICS;

  map<string,Mode>::iterator it;
  for( it = mode_map_.begin(); it != mode_map_.end(); it++ )
    cout << (*it).first << ":" << (*it).second << endl;

	exec_mode_.mode = INVALID;
  read_input();
   
  cout << "mode is: " << exec_mode_.mode << endl;
  switch( exec_mode_.mode )
  {
    case SAMPLE:
      sample_input_data();
			break;
    case LOAD:
      load_data();
			break;
    case PARTS:
      get_parts();
			break;
    case PARTS_ALL:
      get_all_parts();
			break;
    case CLUSTER:
      do_clustering();
			break;
    case EVALUATE:
      do_evaluation();
			break;
    case PART_SEG:
      do_parts_segmentation();
      break;
    case ESTIMATE_REGION_MODELS:
      estimate_region_models();
			break;
		case ESTIMATE_PATCH_THRESHOLD:
			estimate_threshold_protoypes();
			break;
		case COLLECT_STATISTICS:
			collect_statistics();
			break;
    default:
      cout << "unknown operation" << endl;
  }
}


void execute::sample_input_data()
{
  string dir_name = exec_mode_.input_dir;
  dims_meter_ = exec_mode_.dims;
  vector<string> files;
  if( dir_name.find(".pcd") == string::npos ){
    files = get_all_files(dir_name);
    files = sample_data(files,sample_perc_);
  }
}

void execute::load_data()
{
  dims_meter_ = exec_mode_.dims;
  load_cloud(exec_mode_.input_pcd);
}

void execute::get_parts()
{
  string dir_name = exec_mode_.input_dir;
  dims_meter_ = exec_mode_.dims;
  vector<string> files;
  if( dir_name.find(".pcd") == string::npos ){
    files = get_all_files(dir_name);
  }
  else
    files.push_back(dir_name);
  vector<model> models;
  for( size_t i = 0; i < files.size(); i++ )
  {
    cout << "file " << i << " : " << files.size() << endl;
    CloudPtr cloud = load_cloud(files[i]);
    model m = decompose_object(cloud);
    extract_feature_axis(m);
    if( save_ )
      save_feature(m);
    m.id = file_name_;
    models.push_back(m);
  } 
}

void execute::get_all_parts()
{
  patch_ = false;
  string dir_name = exec_mode_.input_dir;
  dims_meter_ = exec_mode_.dims;
  vector<string> files;
  if( dir_name.find(".pcd") == string::npos ){
    files = get_all_files(dir_name);
  }
  else
    files.push_back(dir_name);
  vector<model> models;
  for( size_t i = 0; i < files.size(); i++ )
  {
    cout << "file " << i << " : " << files.size() << endl;
    CloudPtr cloud = load_cloud(files[i]);
    model m = decompose_object(cloud);
    extract_feature_axis(m);
    if( save_ )
      save_feature(m);
    m.id = file_name_;
    models.push_back(m);
  }
}

void execute::estimate_threshold_protoypes()
{
  dims_meter_ = exec_mode_.dims;

  string pro_path = exec_mode_.pro_path;
  cout << "read prototype" << ":" << pro_path <<endl;
	vector<model> pro_models = read_prototypes(pro_path);	
	pair_dists_patch_.clear();
  cout << "refine prototype" << endl;
  for( size_t ip = 0; ip < pro_models.size(); ip++ ){
    cout << "to read" << pro_models[ip].id << endl;
    cout << ip << " from " << pro_models.size() << endl;
    pro_models[ip] = refine_models(pro_models[ip]);
  }

  string file_name;
  file_name = default_path_ + "cls-patch-thresh.txt";
  ofstream ofs(file_name.c_str());
	ofs << pair_dists_patch_[0] << endl;
  ofs.close();
}

void execute::do_clustering()
{
//	agg_max_ = true;
  string dir_name = exec_mode_.input_dir;
  dims_meter_ = exec_mode_.dims;

  string file_name;
  file_name = default_path_ + "cls-patch-thresh.txt";

	double val;
  string line;
  ifstream ifs(file_name.c_str());
  if( ifs.is_open() )
    getline(ifs,line);
  ifs.close();
  stringstream ss(line);
  ss >> val;
	pair_dists_patch_.push_back(val);

  vector<string> files;
  if( dir_name.find(".pcd") == string::npos ){
    files = get_all_files(dir_name);
  }
  else
    files.push_back(dir_name);
  vector<model> models;
  for( size_t i = 0; i < files.size(); i++ )
  {
		cout << "to decompose: " << files[i] << endl;
		cout << i << " from " << files.size() << endl;
    CloudPtr cloud(new Cloud);
		cloud = load_cloud(files[i]);
		if( cloud->points.size() == 0 )
			continue;
    model m = decompose_object(cloud);
    extract_feature_axis(m);
    if( save_ )
      save_feature(m);
    m.id = file_name_;
    models.push_back(m);
  }
  cout << "to cluster" << endl;

  cout << "start cluster: " << models.size() << endl;
  clock_t tf;
  tf = clock();
	
  vector<Hcluster> clusters = cluster_features(models);
	tf = clock() - tf;
  float tf_f = ((float)tf)/CLOCKS_PER_SEC;
  cout << "clustering time : " << tf_f << endl; 	
}


void execute::do_parts_segmentation()
{
  string patch_cls = exec_mode_.patch_cls_mat_path;
  occ_mat_.load(patch_cls);

  cout << "to load patch nn" << endl;
  string patch_n_cls = exec_mode_.patch_n_cls_mat;
  patch_occ_n_.load(patch_n_cls);

  for( int r = 0; r < patch_occ_n_.n_rows;r++)
    patch_occ_n_(r,r) = 0.;

  string dir_name = exec_mode_.patch_cls_path;
  dims_meter_ = exec_mode_.dims;
  parts_cls_ = false;
  vector<string> dirs = get_all_files(dir_name,"cls");
  vector<model> models;
  vector<Hcluster> clusters = read_clusters_patch(dirs,models);
  cout << "patch clusters size: " << clusters.size() << endl;
  cout << "to read prototypes" << endl;
  string pro_path = exec_mode_.pro_path;
  cout << "to evaluate" << endl;
  string eval_path = exec_mode_.eval_path;
  vector<model> eval_models = make_evaluation_models(eval_path);
  double overlap_all = 0;
  int count_p = 0;

  string path_r = default_path_ + "region-thresh.txt";
  ifstream ifs(path_r.c_str());
  if( ifs.is_open() )
  {
    string str;
    if( ifs.good() )
    {
      getline(ifs,str);
      stringstream ss(str);
      ss >> region_threshold_;
    }
  }


  ofstream ofs("overlap-file.txt");

	map<string,int> match_model;
  cout << "after make evaluation models" << endl;
  for( size_t i = 0; i < eval_models.size(); i++ ){
    cout << "to evaluate model patch init" << endl;
    vector<set<int> > regs = segment_parts(eval_models[i],clusters);
    vector<vector<set<int> > > parts_h = parts_h_;
    vector<vector<set<int> > > parts_h_pts = parts_h_pts_;
    make_parts_visualization(parts_h,parts_h_pts,eval_models[i]);
    parts_h_.clear();
    parts_h_pts_.clear();

    vector<CloudPtr> regs_clouds = save_segmented_part(eval_models[i],regs);
    double overlap = estimate_overlap(regs_clouds,eval_models[i]);
		for( size_t ip = 0; ip < p_overlap_.size(); ip++ )
		{
			ofs << eval_models[i].id << " " << p_ids_[ip] << " " << p_overlap_[ip] << endl;
		}
		ofs << eval_models[i].id << " "<< overlap << endl;
    if( overlap < 0 )
      continue;
    count_p++;
    overlap_all += overlap;
  }
	if( (int)count_p > 0)
  	overlap_all /= (double) count_p;
  cout << "number of objs: " << count_p << endl;
  cout << "overall overlap is: " << overlap_all << endl;

	ofs.close();
}

void execute::do_evaluation()
{
  string dir_name = exec_mode_.input_dir;
  dims_meter_ = exec_mode_.dims;
  vector<string> test_data = get_all_files(dir_name);
  map<string,vector<string> > test_map;
  set<string> file_set;
  for( size_t i = 0; i < test_data.size(); i++ )
  {
		cout << "test data: " << test_data[i] << endl;
    string root = test_data[i].substr(test_data[i].find_last_of("/")+1);
    root = root.substr(0,root.find(".pcd"));
    string cue = "_out";
    string sub_root = root.substr(root.find(cue)+cue.length());
    root = root.substr(0,root.find(cue));
    if( sub_root.length() == 0 )
      continue;
    test_map[root].push_back(test_data[i]);
    file_set.insert(root);
  }
  map<int,string> map_id;
  vector<vector<CloudPtr> >clouds;
  map<string,vector<string> >::iterator it;
  for( it = test_map.begin(); it != test_map.end(); it++ )
  {
    vector<CloudPtr> cur_clouds;
    vector<string> vec = (*it).second;
    for( size_t i = 0; i < vec.size(); i++ ){
      CloudPtr cloud = load_cloud(vec[i]);
      cur_clouds.push_back(cloud);
    }
    clouds.push_back(cur_clouds);
    map_id[clouds.size()-1] = (*it).first;
  }
	ofstream ofs("overlap-file.txt");
  cout << "to overlap" << endl;
  double overlap_all = 0;
  double epsilon = 1e-5;
  for( size_t j = 0; j < clouds.size(); j++ )
  {
    model m;
    m.id = map_id[j];
    cout << "pattern: " << map_id[j] << endl;
    double overlap = estimate_overlap(clouds[j],m);
    for( size_t ip = 0; ip < p_overlap_.size(); ip++ )
    {
			ofs << m.id << " " << p_ids_[ip] << " " << p_overlap_[ip] << endl; 
    }
    ofs << m.id << " "<< overlap << endl;
		cout << "obj overlap done" << endl;
    if( overlap > epsilon )
      overlap_all += overlap;
  }
  overlap_all /= (double)clouds.size();
  cout << "overall overlap is: " << overlap_all << endl;
	ofs.close();
}

void execute::collect_statistics()
{
  string dir_name = exec_mode_.patch_cls_path;
  dims_meter_ = exec_mode_.dims;
  cout << "to read clusters" << endl;
  vector<string> dirs = get_all_files(dir_name,"cls");
  vector<model> models;
  vector<Hcluster> clusters = read_clusters_patch(dirs,models);

  cout << "to read prototypes" << endl;
  string pro_path = exec_mode_.pro_path;
  vector<model> pro_models = read_prototypes(pro_path);
  cout << "#of prototypes" << pro_models.size() << endl;

  pair_dists_patch_.clear();

	pair_dists_.clear();
  for( size_t ip = 0; ip < pro_models.size(); ip++ ){
    cout << "id before refine: " << pro_models[ip].id << endl;
		cout << ip << " from : " << pro_models.size() << endl;
    pro_models[ip] = refine_models(pro_models[ip]);
  }


	cout << "to get co-occurance" << endl;
  estimate_co_occurance(models,pro_models,clusters);
  string dir_patch_cls = default_path_ + "patch-cluster/";
  boost::filesystem::path dir_pcls(dir_patch_cls);
  boost::filesystem::create_directories(dir_pcls);
  stringstream ss;
  ss << default_path_ << "patch-cluster/" << "patch-occ";
  occ_mat_.save(ss.str());

	estimate_n_co_occurance(pro_models,clusters);
}

void execute::estimate_region_models()
{
  dims_meter_ = exec_mode_.dims;

	vector<string> dirs_all = get_all_files(exec_mode_.input_dir,"");
	for(size_t i = 0; i < dirs_all.size(); i++)
	{
		string path = dirs_all[i] + "/";
		vector<string> files_cur = get_all_files(path,"_patch");
		vector<string> files_all = get_all_files(path,".pcd");
		cout << "#of patch files: " << files_cur.size() << " : " << files_all.size() << endl;
		string dir_name;	
		for(size_t j = 0; j < files_all.size(); j++)
		{
			if(files_all[j].find("_patch") == string::npos)
				dir_name = files_all[j];
		}	
		cout << "to load: " << dir_name << endl;
		CloudPtr cloud = load_cloud(dir_name);
		vector<CloudPtr> clouds;
		for(size_t j = 0; j < files_cur.size(); j++)
		{
			CloudPtr cloud_patch = load_cloud(files_cur[j]);
			clouds.push_back(cloud_patch);
		}
	
		model m;

		vector<set<int> > surfs_vecs = refine_multiple_patches(clouds,cloud,m);
		for( size_t i = 0; i < surfs_vecs.size(); i++ ){
			stringstream ss;
			ss << m.id << "_" << i;
			file_name_ = ss.str();
			save_prototype(m,surfs_vecs[i]);
		}
	}

}

